
/*
 * $Id: StatPhase.cc,v 1.20.6.6 1999/12/07 02:30:49 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <math.h>
#include <iomanip.h>

#include "Clock.h"
#include "Progress.h"
#include "Connection.h"
#include "Goal.h"
#include "Xaction.h"
#include "ErrorStrTbl.h"
#include "LogComment.h"
#include "PolyOLog.h"
#include "StatPhaseMgr.h"
#include "StatPhase.h"

#include "GoalSym.h"
#include "PhaseSym.h"

#include "globals.h"
#include "polyErrors.h"
#include "polyLogTags.h"


StatPhase::StatPhase(): StatIntvl(theRec), theMgr(0), 
	wasUsed(false), logStats(true),
	theLoadFactBeg(1), theLoadFactEnd(1), theLoadFactK(0),
	theRecurFact(1),
	theLockCount(0) {

	theChannels.append(TheErrChannel);
}

StatPhase::~StatPhase() {
}

void StatPhase::configure(const PhaseSym *cfg, const StatPhase *prevPh) {
	Assert(cfg);

	theRec.theName = cfg->name();
	theGoal.configure(cfg->goal());
	theRec.theGoal = theGoal;

	cfg->logStats(logStats);

	cfg->recurFactor(theRecurFact);

	// use loadFactorEnd from the previous phase if needed
	if (!cfg->loadFactorBeg(theLoadFactBeg) && prevPh)
		theLoadFactBeg = prevPh->theLoadFactEnd;

	// preserve load level unless instructed otherwise
	if (!cfg->loadFactorEnd(theLoadFactEnd))
		theLoadFactEnd = theLoadFactBeg;

	const double fdiff = theLoadFactEnd - theLoadFactBeg;

	// varying load is possible only if goal is specified
	if (fdiff && !theGoal) {
		cerr << cfg->loc() << "phase named `" << theRec.theName
			<< "' has load factors but has no positive goal" << endl;
		exit(-2);
	}

	// cache commonly used value
	// note: duration takes priority over xact count if both are given
	theLoadFactK =
		theGoal.duration() > 0 ?
			fdiff/theGoal.duration().secd() :
		(theGoal.xactCnt() > 0 ?
			fdiff/theGoal.xactCnt() : 0);

	theErrs.maxErrNo(polyErrorStrTbl()->maxErrNo());
}


double StatPhase::loadFactorBeg() const {
	return theLoadFactBeg;
}

double StatPhase::loadFactorEnd() const {
	return theLoadFactEnd;
}

double StatPhase::loadFactor(Time lastFactTm, int lastFactCnt) const {
	const double t = theGoal.duration() > 0 ?
		Max((lastFactTm - theIntvlStart).secd(), 0.0) : (double)lastFactCnt;
	const double lastLF = theLoadFactK*t + theLoadFactBeg;
	const double d = lastLF*lastLF + 4*theLoadFactK;
	const double newLF = d >= 0 ? (sqrt(d) + lastLF)/2 : (lastLF/2);

	// make sure load factor stays between load_fact_beg and load_fact_end
	return theLoadFactBeg <= theLoadFactEnd ?
		MiniMax(theLoadFactBeg, newLF, theLoadFactEnd):
		MiniMax(theLoadFactEnd, newLF, theLoadFactBeg);
}

double StatPhase::recurFactor() const {
	return theRecurFact;
}

void StatPhase::lock() {
	theLockCount++;
}

void StatPhase::unlock() {
	if (locked()) {
		theLockCount--;
		if (!locked())
			checkpoint();
	} else {
		Comment(0) << "internal error: attempt to unlock an unlocked phase (salvaged)" << endl;
	}
}

bool StatPhase::locked() const {
	return theLockCount > 0;
}

Time StatPhase::duration() const {
	return TheClock - theIntvlStart;
}

void StatPhase::start(StatPhaseMgr *aMgr) {
	Assert(!theMgr);
	theMgr = aMgr;

	TheOLog << bege(lgStatPhaseBeg) << name() << ende;

	startListen();
	theIntvlStart = TheClock;

	// manager should be careful with immediate "returns"
	if (!checkpoint()) { 
		// not an immediate exit
		wasUsed = true; 
		if (theGoal.duration() > 0)
			sleepFor(theGoal.duration());
	}
}

void StatPhase::noteConnEvent(BcastChannel *ch, const Connection *c) {
	StatIntvl::noteConnEvent(ch, c);
	if (ch == TheConnCloseChannel) {
		if (!c->bad()) {
			const Time ttl = TheClock - c->openTime();
			theRec.theConnClose.record(c->closeKind(), ttl, c->useCnt());
			theRec.theConnLifeVsUse.record(ttl, c->useCnt());
		}
	}
	checkpoint();
}

void StatPhase::noteXactEvent(BcastChannel *ch, const Xaction *x) {
	StatIntvl::noteXactEvent(ch, x);
	if (ch == TheXactEndChannel) {
		theRec.theRepSzH.record(x->repSize());
		if (x->oidRec().hit)
			theRec.theHitRptmH.record(x->lifeTime());
		else
			theRec.theMissRptmH.record(x->lifeTime());
		theRec.theRptmVsSz.record(x->lifeTime(), x->repSize());
	}
	checkpoint();
}

void StatPhase::noteErrNoEvent(BcastChannel *ch, int errNo) {
	Assert(ch == TheErrChannel);
	theErrs.record(errNo);
}

void StatPhase::name(const String &aName) {
	theRec.theName = aName;
}

void StatPhase::statsLogged(bool are) {
	logStats = are;
}

void StatPhase::wakeUp(const Alarm &alarm) {
	StatIntvl::wakeUp(alarm);
	Assert(locked() || checkpoint());
}

void StatPhase::flush() {
	stop();

	if (logStats) {
		TheOLog 
			<< bege(lgErrorStat) << (Array<int>&)theErrs << ende
			<< bege(lgStatPhaseRec) << theRec << ende
			;
	}
}

bool StatPhase::checkpoint()  {
	if (!locked() && theGoal.reached(*this)) {
		stop();
		report();

		Assert(theMgr);
		StatPhaseMgr *mgr = theMgr;
		theMgr = 0;
		mgr->noteDone(this);
		return true;
	}
	return false;
}

void StatPhase::stop() {
	Assert(theIntvlStart >= 0);
	if (theRec.theDuration >= 0) // already stopped
		return;

	theRec.theDuration = TheClock - theIntvlStart;
	if (theRec.theDuration < 0) // sometimes clock goes backwards!
		theRec.theDuration = Time(0,0);

	stopListen();
	TheAlarmClock.cancelAll(this);

	TheOLog << bege(lgStatPhaseEnd) << name() << ende;
}

void StatPhase::report() const {
	Comment(1) << "p-" << name();
	StatIntvl::report(Comment);
	Comment << endc;
}


void StatPhase::reportCfg(ostream &os) const {
	os 
		<< setw(10) << name()
		<< ' ' << setw(8) << theLoadFactBeg 
		<< ' ' << setw(8) << theLoadFactEnd
		<< ' ' << setw(5) << theGoal
		<< "\t " << (logStats ? "" : "!log")
		<< endl;
}
