
/*
 * $Id: Xaction.cc,v 1.30.2.4 1999/12/06 05:16:10 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <ctype.h>

#include "Agent.h"
#include "Connection.h"
#include "Clock.h"
#include "Progress.h"
#include "ErrorMgr.h"
#include "SharedOpts.h"
#include "PolyOLog.h"
#include "XactStatRec.h"
#include "Xaction.h"

#include "polyBcastChannels.h"
#include "polyErrors.h"
#include "globals.h"

int Xaction::TheSampleDebt = 0;


void Xaction::reset() {
	theConn = 0;
	theId.clear();
	theStartTime = theLifeTime = Time();
	theRepSize = -1;
	theOidRec.reset();
	theEmbedCount = 0;
	theState = stNone;
}

void Xaction::start(Connection *aConn) {
	Assert(aConn);
	theConn = aConn;

	theStartTime = theConn->useStart();
	Broadcast(TheXactBegChannel, this);
}

void Xaction::stop(int errNo) {
	Assert(theConn);
	if (theState == stSpaceWaiting)
		theConn->theWr.stop(this);
	else
		theConn->theRd.stop(this);
	
	theLifeTime = TheClock - theStartTime;

	if (!errNo) {
		stopStats();
		if (TheSampleDebt) {
			TheSampleDebt--;
			logStats();
		}
	} else {
		theConn->bad(true);
		// low level errors should be reported when they occur
		// low level errors are propogated to us as "-1"
		// do not count/report them twice
		if (errNo > 0) {
			// needExpl counts the error and print err msg if needed
			if (ReportError(errNo) && TheOpts.theDumpFlags(dumpErr, dumpAny))
				printMsg(theConn->theRdBuf);
		}
		Broadcast(TheXactErrChannel, this);
		TheProgress.failure();
	}

	theState = stDone;
}

void Xaction::stopStats() {
	TheProgress.success();
	Broadcast(TheXactEndChannel, this);
}

void Xaction::logStats() {
	XactStatRec xr;

	xr.theId = theId;
	xr.theOwnerSeqId = owner()->seqId();
	xr.theConnSeqId = theConn->seqId();
	xr.times(theStartTime, theEnqueTime, theLifeTime);
	xr.theEmbedCount = theEmbedCount;
	xr.theRepSize = theRepSize;
	xr.theOidRec = theOidRec;

	xr.store(*TheSmplOLog);
}

void Xaction::printMsg(const IOBuf &buf) const {
	printMsg(buf, buf.contSize());
}

void Xaction::printMsg(const IOBuf &buf, Size maxSize) const {
	if (buf.contSize() <= 0)
		printMsg(0, 0);
	else
		printMsg(buf.content(), maxSize);
}

inline
bool myIsPrint(char c) {
	return (c && isprint(c)) || c == '\r' || c == '\n';
}

// print buf content up to the first non-printable character,
// and no more than maxSize characters
void Xaction::printMsg(const char *buf, Size maxSize) const {
	if (theId) {
		cout << TheClock << '#';
		theOidRec.print(cout << " obj: ");
		cout << " xact: " << theId << endl;
	}

	if (!buf) {
		cout << "[no data to dump]" << endl;
		return;
	}

	if (maxSize > TheOpts.theDumpSize)
		maxSize = TheOpts.theDumpSize;

	const char *p = buf;
	while ((p-buf) < maxSize && myIsPrint(*p)) ++p;

	if (p > buf) {
		cout.write(buf, p - buf);
		cout << endl;
	}
}
