
/*
 * $Id: aka.cc,v 1.1.2.8 1999/10/20 18:53:48 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <arpa/inet.h>

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#include "CmdLine.h"
#include "opts.h"

#include "PglNetAddrRange.h"
#include "getIfAddrs.h"
#include "gadgets.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",         "list of options"),
			theIfName(this,      "if <str>",     "interface name"),
			theAlias(this,       "aliases <str>","ip or an ip range to set as an alias")
			{}

		virtual bool validate() const;

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		StrOpt theIfName;
		StrOpt theAlias;
};


struct ifaliasreq_in {
	char    ifra_name[IFNAMSIZ];   /* if name, e.g. "en0" */
	struct  sockaddr_in ifra_addr;
	struct  sockaddr_in ifra_broadaddr;
	struct  sockaddr_in ifra_mask;
};


static MyOpts TheOpts;
static PglNetAddrRange TheAliasParser;
static Array<in_addr>  TheOldAliases;
static Array<NetAddr*> TheAliases;
static NetAddr ThePrimaryAddr;
static NetAddr ThePrimaryBcastAddr;


/* MyOpt */

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "[interface_name] [alias]";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	if (!opts.count() || opts.count() > 2)
		return false;

	if (!isdigit(*opts[0])) {
		theIfName.val(opts[0]);
		if (opts.count() == 2)
			theAlias.val(opts[1]);
	} else {
		theAlias.val(opts[0]);
	}

	return true;
}

bool MyOpts::validate() const {
	if (!theIfName)
		cerr << "interface name is not specified" << endl;
	else
	if (theAlias && !TheAliasParser.parse(theAlias))
		cerr << "malformed alias format: `" << theAlias << "'" << endl;
	else
		return true;
	return false;
}

static
void configureLogs(int prec) {
	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

static
void configure() {
	configureLogs(2);

	TheAliasParser.toAddrs(TheAliases);
}

static
void getPrimaryAddress() {
	sockaddr_in addr;
	Socket s;
	Must(s.create(AF_INET, SOCK_DGRAM, 0));
	Must(s.getIfAddr(TheOpts.theIfName, (sockaddr&)addr));
	ThePrimaryAddr = NetAddr(addr);
	Must(s.getIfBcastAddr(TheOpts.theIfName, (sockaddr&)addr));
	ThePrimaryBcastAddr = NetAddr(addr);
	s.close();

	ThePrimaryAddr.port(-1);
	ThePrimaryBcastAddr.port(-1);
	clog << TheOpts.theIfName << ": primary address is " << ThePrimaryAddr << " broadcast " << ThePrimaryBcastAddr << endl;
}

static
void getOldAliases() {
	Must(GetIfAddrs(TheOldAliases, TheOpts.theIfName));

	// remove primary address from the aliases array
	bool found = false;
	for (int i = 0; !found && i < TheOldAliases.count(); ++i) {
		if (TheOldAliases[i] == ThePrimaryAddr.addrN()) {
			TheOldAliases[i] = TheOldAliases.last();
			TheOldAliases.pop();
			found = true;
		}
	}

	if (!found)
		clog << TheOpts.theIfName << ": hmm.. primary address is not"
			<< " in the interface alias list." << endl;

	clog << TheOpts.theIfName << ": found "
		<< TheOldAliases.count() << " old aliases." << endl;
}

static
void delOldAliases() {
	Socket s;
	Must(s.create(AF_INET, SOCK_DGRAM, 0));
	for (int i = 0; i < TheOldAliases.count(); ++i) {
		struct ifaliasreq_in r;
		memset(&r, 0, sizeof(r));
		strncpy(r.ifra_name, TheOpts.theIfName, sizeof(r.ifra_name));
// XXX: remove all these defines
#ifdef HAVE_SA_LEN
		r.ifra_addr.sin_len = sizeof(r.ifra_addr);
#endif
		r.ifra_addr.sin_family = AF_INET;
		r.ifra_addr.sin_port = 0;
		r.ifra_addr.sin_addr = TheOldAliases[i];
		Must(s.delIfAddr((ifaliasreq&)r));
	}
	s.close();
	clog << TheOpts.theIfName << ": deleted "
		<< TheOldAliases.count() << " old aliases." << endl;
}

static
void setNewAliases() {
	Socket s;
	Must(s.create(AF_INET, SOCK_DGRAM, 0));
	for (int i = 0; i < TheAliases.count(); ++i) {
		struct ifaliasreq_in r;
		memset(&r, 0, sizeof(r));
		strncpy(r.ifra_name, TheOpts.theIfName, sizeof(r.ifra_name));
#ifdef HAVE_SA_LEN
		r.ifra_addr.sin_len = sizeof(r.ifra_addr);
#endif
		r.ifra_addr.sin_family = AF_INET;
		r.ifra_addr.sin_port = 0;
		r.ifra_addr.sin_addr = TheAliases[i]->addrN();

		// set network mask (geting old netmask is not documented?!)
		r.ifra_mask = r.ifra_addr;
		TheAliasParser.netmask(r.ifra_mask.sin_addr);

		// set broacast address
		r.ifra_broadaddr = r.ifra_addr;
		r.ifra_broadaddr.sin_addr.s_addr |= ~(unsigned int)r.ifra_mask.sin_addr.s_addr;
		//r.ifra_broadaddr.sin_addr = ThePrimaryBcastAddr.addrN();

		if (!s.addIfAddr((ifaliasreq&)r) && errno != EEXIST)
			Must(false);
	}
	s.close();
	clog << TheOpts.theIfName << ": created "
		<< TheAliases.count() << " new aliases." << endl;
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	clog << TheOpts.theIfName << ": ";
	if (TheOpts.theAlias)
		clog << "reseting aliases to " << TheOpts.theAlias << endl;
	else
		clog << "no new alias specified; will just delete old ones." << endl;

	getPrimaryAddress();
	getOldAliases();
	delOldAliases();

	if (TheOpts.theAlias)
		setNewAliases();

	return 0;
}
