
/*
 * $Id: lx.cc,v 1.23.4.9 1999/12/15 15:21:28 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <limits.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "CmdLine.h"
#include "opts.h"

#include "StatPhaseRec.h"
#include "ILog.h"
#include "DLog.h"
#include "IntvlLx.h"
#include "PhaseLx.h"
#include "ErrorLx.h"
#include "polyLogTags.h"
#include "ilog2dlog.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",         "list of options"),
			theOutFileName(this, "out <file>",   "redirect console output", "-"),
			thePhaseName(this,   "phase <str>",  "phase name", "All"),
			theObjects(this,     "objects <str>","names of objects to extract", "All"),
			theTrace(this,       "trace <time>",  "print (averaged at specified intervals) run-time stats (rather than totals)"),
			theTimeScale(this,   "time_scale <time>", "adjust time column in traces"),
			printTime(this,      "prn_time <bool>", "prepend lines with time since run start"),
			printRepCnt(this,    "prn_rep_cnt <bool>", "prepend lines with number of replies received"),
			printNames(this,     "prn_names <bool>", "prepend object values with object names", true)
			{}

		virtual bool validate() const;

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	protected:
		void planExtor(LogExtractor *e) const;

	public:
		HelpOpt theHelpOpt;
		StrOpt theOutFileName;
		StrOpt thePhaseName;
		StrOpt theObjects;
		TimeOpt theTrace;
		TimeOpt theTimeScale;
		BoolOpt printTime;
		BoolOpt printRepCnt;
		BoolOpt printNames;

		Array<String*> theFileNames;
};

static String getTraceables();
static LogExtractor *getExtor(const String &name, const LxParams &params);

static MyOpts TheOpts;
static DLog TheDLog;
static SchArray<LogExtractor*> ThePlan;

// `All' macro
static const String TheNames = 
	"req_rate,rep_rate,"
	"conn_ttl_mean,conn_use_mean,req_dhr,req_bhr,"
	"rep_dhr,rep_bhr,rep_dcr,rep_bcr,"
	"rptm_mean,rptm_mean:hit,rptm_mean:miss,"
	"rep_sz_mean,fill_sz,fill_cnt,"
	"err_cnt,err_pcnt,duration,"
	"phase_name,rptm_hist,"
	"conn_use_hist,conn_ttl_hist,"
	"conn_close_act_use_hist,conn_close_act_ttl_hist,"
	"conn_close_frgn_use_hist,conn_close_frgn_ttl_hist,"
	"conn_close_lcl_use_hist,conn_close_lcl_ttl_hist,"
	"soread_sz_hist,sowrite_sz_hist,rep_sz_hist,"
	"rptm_vs_sz,conn_ttl_vs_use,"
	"err_cnt_tbl"
	;
// hidden: conn_lvl_mean

static const String TheTraceableNames = getTraceables();


/* MyOpt */

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "<log_file_name> ...";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	for (int i = 0 ; i < opts.count(); ++i)
		theFileNames.append(new String(opts[i]));

	// add default fname if none are specified
	if (!theFileNames.count())
		theFileNames.append(new String("-"));

	return theFileNames.count() > 0;
}

bool MyOpts::validate() const {
	// convert object names into extractors
	// XXX: misplaced, should be done in a custom Opt
	String buf = TheOpts.theObjects;
	while (buf) {
		String head, tail;
		if (splitVal(buf, head, tail, ','))
			buf = tail;
		else
			head = buf, buf = 0;

		String name, paramStr;
		if (!splitVal(head, name, paramStr, ':'))
			paramStr = 0;

		LxParams params(paramStr);

		if (name == "Traces") {
			if (params.count()) {
				cerr << "`Traces' macro cannot have parameters; got `" << params.image() << "'" << endl;
				return false;
			}
			if (TheOpts.theTrace < 0) {
				cerr << "`Traces' macro must be used with `--trace' option" << endl;
				return false;
			}
			buf = buf ? TheTraceableNames + "," + buf : TheTraceableNames;
			continue;
		}

		if (name == "All") {
			if (params.count()) {
				cerr << "`All' macro cannot have parameters; got `" << params.image() << "'" << endl;
				return false;
			}
			buf = buf ? TheNames + "," + buf : TheNames;
			continue;
		}

		if (LogExtractor *extor = getExtor(name, params)) {
			planExtor(extor);
		} else {
			cerr << "unknown object name `" << head << "'" << endl;
			return false;
		}
	}

	if (!ThePlan.count())
		cerr << "no objects specified" << endl;
	else
	if (theTimeScale > 0 && (theTrace < 0 || !printTime))
		cerr << "time scale requires enabled --trace and --prn_time options" << endl;
	else
		return true;
	return false;
}

void MyOpts::planExtor(LogExtractor *e) const {
	ThePlan.append(e);
}


/* local routines */

static
void printTime(Time tm) {
	static Time start = tm;
	if (TheOpts.theTimeScale > 0) {
		tm -= start;
		cout << (tm/TheOpts.theTimeScale);
	} else {
		cout << tm.secd();
	}
	cout << ' ';
}

// extract requested objects (both ping and soft)
static
void extract(const Array<LogExtractor*> &plan) {

	if (TheOpts.printTime && !TheOpts.printNames)
		printTime(TheDLog.theProgress.time());

	if (TheOpts.printRepCnt && !TheOpts.printNames)
		cout << TheDLog.theProgress.xacts() << ' ';

	const LxRec &r = TheOpts.theTrace > 0 ? 
		(LxRec &)TheDLog.theCycleLxRec : (LxRec &)TheDLog.thePhaseLxRec;

	// extract actual values
	for (int i = 0; i < plan.count(); ++i) {

		if (TheOpts.printNames) {
			if (TheOpts.printTime)
				printTime(TheDLog.theProgress.time());
			if (TheOpts.printRepCnt)
				cout << TheDLog.theProgress.xacts() << ' ';
			plan[i]->label(cout);
			cout << ":\t ";
		} else
		if (i)
			cout << ' '; // insert delimiter

		if (plan[i]->name() == "err_cnt_tbl") // XXX: kludge
			plan[i]->extract(TheDLog.theErrorLxRec, cout);
		else
			plan[i]->extract(r, cout);

		if (TheOpts.printNames)
			cout << endl;
	}

	if (!TheOpts.printNames)
		cout << endl;
}

static void extractTrace(const DLog &) { extract(ThePlan); }

static 
String getTraceables() {
	String res = "";
	String buf = TheNames;
	while (buf) {
		String head, tail;
		if (splitVal(buf, head, tail, ','))
			buf = tail;
		else
			head = buf, buf = 0;

		String name, paramStr;
		if (!splitVal(head, name, paramStr, ':'))
			paramStr = 0;

		LxParams params(paramStr);
		IntvlLx ilx;
		if (ilx.configure(name, params)) {
			if (res.len())
				res += ",";
			res += name;
		}
	}
	return res;
}

static 
LogExtractor *getExtor(const String &name, const LxParams &params) {
	LogExtractor *lx = 0;
	PhaseLx plx;
	ErrorLx elx;
	if (plx.configure(name, params)) {
		if (TheOpts.theTrace < 0) {
			lx = new PhaseLx;
			Assert(lx->configure(name, params));
		} else {
			lx = new IntvlLx;
			if (!lx->configure(name, params)) {
				cerr << "object `" << name << "' is known but not traceable" << endl;
				exit(-2);
			}
		}
	} else
	if (elx.configure(name, params)) {
		if (TheOpts.theTrace >= 0) {
			cerr << "object `" << name << "' is known but not traceable" << endl;
			exit(-2);
		}
		lx = new ErrorLx;
		Assert(lx->configure(name, params));
	}
	return lx;
}

static
void configureLogs(int prec) {
	if (TheOpts.theOutFileName && TheOpts.theOutFileName != "-")
		cout = *new ofstream(TheOpts.theOutFileName);

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

static
void configure() {
	configureLogs(2);

	if (TheOpts.theTrace > 0) {
		TheDLog.theCutter.configure(TheOpts.theTrace);
		TheDLog.filter(&extractTrace);
	}

	if (TheOpts.thePhaseName == "All")
		TheDLog.thePhaseName = 0; // no restrictions
	else
		TheDLog.thePhaseName = TheOpts.thePhaseName;
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	// print all requested names unless we will do it later
	if (!TheOpts.printNames) {
		cout << "#objects:";
		for (int i = 0; i < ThePlan.count(); ++i)
			ThePlan[i]->label(cout << ' ');
		cout << endl;
	}

	// suck all data in
	for (int i = 0; i < TheOpts.theFileNames.count(); ++i) {
		if (!ilog2dlog(*TheOpts.theFileNames[i], TheDLog))
			cerr << here << "error processing log " << *TheOpts.theFileNames[i] << endl;
	}

	if (TheOpts.theTrace <= 0)
		extract(ThePlan);

	return 0;
}
